#
# main.py
#

# fun stuff to use
import time
import random
import math
import sys

# here's the base class, hiding details...
from ThreadedClient import *

# not sure that this needs to be a class, but...
class RobotController(ThreadedClient):
    """
    A class defining the graphical display and the behavior of
    a Roomba-like robot (except that it has a range sensor!)
    """
    
    def __init__(self, robotType, mapname):
        """
        This constructor delegates things to the ThreadedClient class
        """
        ThreadedClient.__init__(self, robotType, mapname)
        
    
    def main(self):
        """
        This while loop is, in essence, the life of the robot.
        Each time through the loop, you will want to
           * read the sensors
           * decide on what action to take (if any)
           * execute the current action
        Then, it all begins again!
        """
        
        
        # pause for a second
        time.sleep(1)
        
        # these constants represent different states in our state machine
        STOP = 0
        KBD = 1
        TURN = 2
        GO = 3
        WANDER = 4
        robotTask = KBD      # this is the starting state
        
        # other useful constants
        THRESH = 16.0        # radius of the roomba, in cm
        noise = 0.05
        
        # an example of how to get the map here in the controller
        segList = self.getMap()
        counter = 0
        for s in segList:
            #print 'segment #', counter, 'is (', s.x1, ',', s.y1, ') - (', s.x2, ',', s.y2, ')'
            counter += 1
        
        rangeHeading = 0.0   # where to point our range finder
        
        oldgoal = (0,0)      # the old goal
        goal = (0,0)         # the current goal
        
        obstacles = []       # add points to this...
        
        while True:          # run until you hit 'Q' in the window
                             # or control-c at the command prompt
            
            # SENSE
            #
            # x is the x coordinate of the robot
            # y is the y coordinate of the robot
            # thd is the robot's heading, theta, in degrees
            # bump is a list of two bump sensor readings: left and right
            [x,y,thd], bump = self.getData()
            #print 'x,y,thd:', x, y, thd
            # get vision data here...
            
            # get the range to the walls at a particular heading
            d = self.getRange(rangeHeading)     # in degrees
            
            # thr is the robot's heading, theta, in radians
            thr = math.radians(thd)             # nice of python to have this function
            
            # get any keypresses and goal locations (mouse clicks when 'g' is pressed)
            kbd, goal = self.getWindowEvents()  # at start, kbd == '' and goal == (0,0)
            
            if goal != oldgoal:
                # a new goal has been clicked - let's confirm this on the screen
                print 'New goal at x =', goal[0], 'and y =', goal[1]
                self.showGoal( goal )           # displays the goal in green
                oldgoal = goal                  # reset the old goal
            
            # have we gotten to the goal?
            if robotTask != STOP and robotTask != KBD:
                [realx, realy, realthr] = self.getRealPose()
                distToGoal = math.sqrt( (realx-goal[0])**2 + (realy-goal[1])**2 )
                if distToGoal < THRESH:
                    print 'At goal!'
                    self.setVels(0,0)
                    robotTask = KBD
            
            
            # PLAN
			#
            # This stops if a bump is encountered - a pretty basic plan
            if bump[0] == True or bump[1] == True:
                print 'BUMP!',
                print '  [Left bump sensor:',  bump[0], ']  ',
                print '  [Right bump sensor:', bump[1], ']  '
                robotTask = STOP
            
            
            # ACT
            #
            # here, we decide how to set velocities,
            # based on our current state
            if robotTask == GO:
                self.setVels(40,0,noise)
                robotTask = KBD     # we've stopped, so let's return
            
            if robotTask == STOP:
                self.setVels(0,0)   # stop!
                robotTask = KBD     # we've stopped, so let's return
            #
            # the KBD state is for user-control
            #
            if robotTask == KBD:
                # We allow keypresses to override or set our actions
                #
                if kbd == 'i': self.setVels(25,0,noise)
                elif kbd == 'k': self.setVels(-25,0,noise)
                elif kbd == 'j': self.setVels(10,40,noise)
                elif kbd == 'l': self.setVels(10,-40,noise)
                elif kbd == 'P':
                    # this call creates the graphical representations of one particle
                    particles = [ [random.uniform(-100,100),0,0,1.0,0] ];
                    self.queue.put([self.gui.setParticles, particles])
                
                elif kbd == 'M':
                    print 'Move'
                
                elif kbd == 'S':
                    print 'Sense'
                    
                elif kbd == 'w':
                    print 'wall!'
                    # add a wall point here
                    obstacles.append( (x,y) )
                    self.queue.put([self.gui.addLine, x, y, 0.0])
                    print 'obstacles is', obstacles
                
                elif kbd == 'R':
                    print 'Resample'
                
                elif kbd == 'G':
                    print 'toggle MCL'
                
                elif kbd == ' ':
                    robotTask = KBD
                    self.setVels(0,0)
            
            # special case to allow us to quit
            if kbd == 'Q': break              # 'Q' quits the program

            
            # We pause to let the graphics catch up...
            time.sleep(0.025) # 1/40 of a second
            
        
        # this is outside the while loop...
        # shutdown the robot and gui, if needed
        self.shutdown()
            


# this is python's way of detecting if you've started this file from
# the command line and, if so, instantiating the above class and
# starting the run() method...
#
if __name__ == '__main__':      # are we at the command line?
    
    # our pre-provided array of maps...
    MAPS = [ 'mapBlank.txt', 
             'map0.txt',
             'map1.txt',
             'map2rooms.txt',
             'mapColors.txt',
             'mapMaze.txt',
             'mapHallways.txt',
             'mapLibraComplex.txt' ]
    
    # give it a map name and a robot type...
    #R = RobotController(ThreadedClient.REAL_ROOMBA, MAPS[1])  # create R
    R = RobotController(ThreadedClient.SIM_ROOMBA, MAPS[0])  # create R
    R.run()                     # run R

    
    
